<?php
declare(strict_types=1);

namespace Infomaker\Everyware\Support;

use ArrayAccess;
use Illuminate\Contracts\Support\Arrayable;
use Illuminate\Contracts\Support\Jsonable;
use JsonSerializable;
use Infomaker\Everyware\Support\Interfaces\PropertyObject;

/**
 * GenericPropertyObject
 *
 * @link    http://infomaker.se
 * @package Infomaker\Everyware\Base
 * @since   Infomaker\Everyware\Support\GenericPropertyObject 1.0.0
 */
class GenericPropertyObject implements ArrayAccess, Arrayable, Jsonable, JsonSerializable, PropertyObject
{

    /**
     * All of the properties set on the container.
     *
     * @var array
     */
    protected $properties = [];

    /**
     * Fill up properties dynamically
     *
     * @param array $properties
     *
     * @since 0.1
     * @return void
     */
    public function fill(array $properties = []): void
    {
        foreach ($properties as $key => $value) {
            $this->set($key, $value);
        }
    }

    /**
     * Get an property from the container.
     *
     * @param string $key
     * @param mixed  $default
     *
     * @since 0.1
     * @return mixed
     */
    public function get($key, $default = null)
    {
        if (array_key_exists($key, $this->properties)) {
            return $this->properties[$key];
        }

        return value($default);
    }

    /**
     * @param $key
     *
     * @return bool
     */
    public function getBoolValue($key): bool
    {
        return filter_var($this->getSingleValue($key), FILTER_VALIDATE_BOOLEAN);
    }

    /**
     * Retrieve a date version of a value
     *
     * @param $key
     *
     * @since 1.0.0
     * @return Date
     */
    public function getDateValue($key): Date
    {
        return Date::createFromOcString($this->getSingleValue($key));
    }

    /**
     * Get property as multi value (array)
     *
     * @param string $property
     *
     * @since 0.1
     * @return array
     */
    public function getMultiValue($property): array
    {
        return $this->offsetExists($property) ? (array)$this->get($property) : [];
    }

    /**
     * Get property as a single value
     *
     * @param string $property
     *
     * @since 0.1
     * @return mixed
     */
    public function getSingleValue($property)
    {
        $value = $this->getMultiValue($property);

        return ! empty($value) ? array_shift($value) : '';
    }

    /**
     * Get first of single value
     *
     * @param $selectors
     *
     * @return string
     */
    public function getFirstOfSingleValue($selectors): string
    {
        return $this->getFirstOf($selectors, true);
    }

    /**
     * Get first of multi value
     *
     * @param $selectors
     *
     * @return array
     */
    public function getFirstOfMultiValue($selectors): array
    {
        return $this->getFirstOf($selectors, false);
    }

    /**
     * Get first of property
     *
     * @param      $selectors
     * @param bool $single_value
     *
     * @return array|string
     */
    public function getFirstOf($selectors, $single_value = true)
    {
        $selectors = \is_array($selectors) ? $selectors : $this->parseSelectors($selectors);
        foreach ($selectors as $selector) {
            $value = $single_value ? $this->getSingleValue($selector) : $this->getMultiValue($selector);
            if ( ! empty($value)) {
                return $value;
            }
        }

        return $single_value ? '' : [];
    }

    /**
     * Parse string of selectors into an array
     * Multiple selector may come separated by space, comma or pipe
     *
     * @param string $selectors
     *
     * @since 0.1
     * @return array
     */
    protected function parseSelectors($selectors = ''): array
    {
        return (array)preg_split('/\s*(,|\|)\s*/', $selectors);
    }

    /**
     * Get an property from the container.
     *
     * @param string $key
     * @param mixed  $value
     *
     * @since 0.1
     * @return void
     */
    public function set($key, $value): void
    {
        $this->properties[$key] = $value;
    }

    /**
     * Get the properties from the container.
     *
     * @since 0.1
     * @return array
     */
    public function getProperties(): array
    {
        return $this->properties;
    }

    /**
     * Convert the Fluent instance to an array.
     *
     * @since 0.1
     * @return array
     */
    public function toArray(): array
    {
        return array_map(function ($value) {
            return $value instanceof Arrayable ? $value->toArray() : $value;
        }, $this->properties);
    }

    /**
     * Convert the object into something JSON serializable.
     *
     * @since 0.1
     * @return array
     */
    public function jsonSerialize(): array
    {
        return $this->toArray();
    }

    /**
     * Convert the Fluent instance to JSON.
     *
     * @param int $options
     *
     * @since 0.1
     * @return string
     */
    public function toJson($options = 0): string
    {
        return json_encode($this->jsonSerialize(), $options);
    }

    /**
     * Determine if the given offset exists.
     *
     * @param string $offset
     *
     * @since 0.1
     * @return bool
     */
    public function offsetExists($offset): bool
    {
        return isset($this->{$offset});
    }

    /**
     * Get the value for a given offset.
     *
     * @param string $offset
     *
     * @since 0.1
     * @return mixed
     */
    public function offsetGet($offset)
    {
        return $this->{$offset};
    }

    /**
     * Set the value at the given offset.
     *
     * @param string $offset
     * @param mixed  $value
     *
     * @since 0.1
     * @return void
     */
    public function offsetSet($offset, $value): void
    {
        $this->{$offset} = $value;
    }

    /**
     * Unset the value at the given offset.
     *
     * @param string $offset
     *
     * @since 0.1
     * @return void
     */
    public function offsetUnset($offset): void
    {
        unset($this->{$offset});
    }

    /**
     * Handle dynamic calls to the container to set properties.
     *
     * @param string $method
     * @param array  $parameters
     *
     * @since 0.1
     * @return $this
     */
    public function __call($method, $parameters)
    {
        $this->properties[$method] = \count($parameters) > 0 ? $parameters[0] : true;

        return $this;
    }

    /**
     * Dynamically retrieve the value of an property.
     *
     * @param string $key
     *
     * @since 0.1
     * @return mixed
     */
    public function __get($key)
    {
        return $this->get($key);
    }

    /**
     * Dynamically set the value of an property.
     *
     * @param string $key
     * @param mixed  $value
     *
     * @since 0.1
     * @return void
     */
    public function __set($key, $value)
    {
        $this->set($key, $value);
    }

    /**
     * Dynamically check if an property is set.
     *
     * @param string $key
     *
     * @since 0.1
     * @return bool
     */
    public function __isset($key)
    {
        return isset($this->properties[$key]);
    }

    /**
     * Dynamically unset an property.
     *
     * @param string $key
     *
     * @since 0.1
     * @return void
     */
    public function __unset($key)
    {
        unset($this->properties[$key]);
    }
}
