<?php declare(strict_types=1);

namespace Infomaker\Everyware\Support\Interfaces;

interface PropertyObject {

    /**
     * Fill up properties dynamically
     *
     * @param array $properties
     *
     * @since 0.1
     * @return void
     */
    public function fill( array $properties = [] ): void;

    /**
     * Get an property from the container.
     *
     * @param string $key
     * @param mixed  $default
     *
     * @since 0.1
     * @return mixed
     */
    public function get( string $key, $default = null );

    /**
     * Get an property from the container.
     *
     * @param string $key
     * @param mixed  $value
     *
     * @since 0.1
     * @return void
     */
    public function set( string $key, $value );

    /**
     * Get the properties from the container.
     *
     * @since 0.1
     * @return array
     */
    public function getProperties(): array;

    /**
     * Convert the Fluent instance to an array.
     *
     * @since 0.1
     * @return array
     */
    public function toArray(): array;

    /**
     * Convert the object into something JSON serializable.
     *
     * @since 0.1
     * @return array
     */
    public function jsonSerialize(): array;

    /**
     * Convert the Fluent instance to JSON.
     *
     * @param int $options
     *
     * @since 0.1
     * @return string
     */
    public function toJson( int $options = 0 ): string;

    /**
     * Determine if the given offset exists.
     *
     * @param string $offset
     *
     * @since 0.1
     * @return bool
     */
    public function offsetExists( string $offset ): bool;

    /**
     * Get the value for a given offset.
     *
     * @param string $offset
     *
     * @since 0.1
     * @return mixed
     */
    public function offsetGet( string $offset );

    /**
     * Set the value at the given offset.
     *
     * @param string $offset
     * @param mixed  $value
     *
     * @since 0.1
     * @return void
     */
    public function offsetSet( string $offset, $value ): void;

    /**
     * Unset the value at the given offset.
     *
     * @param string $offset
     *
     * @since 0.1
     * @return void
     */
    public function offsetUnset( string $offset ): void;

    /**
     * Handle dynamic calls to the container to set properties.
     *
     * @param string $method
     * @param array  $parameters
     *
     * @since 0.1
     * @return $this
     */
    public function __call( string $method, array $parameters );

    /**
     * Dynamically retrieve the value of a property.
     *
     * @param string $key
     *
     * @since 0.1
     * @return mixed
     */
    public function __get( string $key );

    /**
     * Dynamically set the value of a property.
     *
     * @param string $key
     * @param mixed  $value
     *
     * @since 0.1
     * @return void
     */
    public function __set( string $key, $value );

    /**
     * Dynamically check if a property is set.
     *
     * @param string $key
     *
     * @since 0.1
     * @return bool
     */
    public function __isset( string $key );

    /**
     * Dynamically unset a property.
     *
     * @param string $key
     *
     * @since 0.1
     * @return void
     */
    public function __unset( string $key );

}
